<?php
require_once(ABSPATH . 'wp-admin/includes/media.php');
require_once(ABSPATH . 'wp-admin/includes/file.php');
require_once(ABSPATH . 'wp-admin/includes/image.php');

class SWRetail{
    /**
     * SW-Retail plugin licence key
     * @var string
     */
    protected $swLicenceKey = '';

    /** 
     * SW-Retail URL
     * @var string
     */
    protected $swURL = '';

    /**
     * SW-Retail Username
     * @var string
     */
    protected $swUser = '';

    /**
     * SW-Retail Password
     * @var string
     */
    protected $swPassword = '';

    /**
     * WP remote get/post args
     * @var array
     */
    protected $wpRemoteArgs = [];

    /**
     * SWRetail constructor.
     *
     * @param $swLicenceKey
     * @param $swURL
     * @param $swUser
     * @param $swPassword
     */
    public function __construct($swLicenceKey, $swURL, $swUser, $swPassword){
        $this->swLicenceKey	= $swLicenceKey;
        $this->swURL 		= $swURL;
        $this->swUser 		= $swUser;
        $this->swPassword 	= $swPassword;
        $this->wpRemoteArgs = [
            'headers' => [
                'Authorization' => 'Basic ' . base64_encode( $this->swUser . ':' . $this->swPassword),
                'X-SW-PARTNER-MULL2MEDIA' => 1
            ],
            'timeout'     => 10,
        ];
    }

    /**
     * Create task to execute at a later time.
     *
     * @param string $action
     * @param string $params
     */
    public function createSwTask($action, $params = ''){
        global $wpdb;
        $query = ['action' => $action];
        if(!empty($params)){
            $query['params'] = $params;
        }
        $wpdb->insert('swretail_tasks', [
            'task' 			=> site_url().'/api/apiservice.php?'.http_build_query($query),
            'date_added' 	=> date('Y-m-d H:i:s'),
        ]);
    }

    /**
     * Test SW-retail connection.
     */
    public function testSwConnection(){
        if(!$this->checkLicence()){return;}
        try{
            swretailUpdateApiRequestCount('conn_test');
            $response = wp_remote_get( $this->swURL.'conn_test', $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', testSwConnection()');
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode)){
                    ?>
                    <div class="notice notice-success is-dismissible">
                        <p><?php _e('Connection successful!', 'swretail'); ?></p>
                    </div>
                    <?php
                }else{
                    ?>
                    <div class="notice notice-error is-dismissible">
                        <p><?php _e('Error connection failed:', 'swretail'); echo ' #'.$response->errorcode.' - '.$response->errorstring; ?></p>
                    </div>
                    <?php
                }
            }
        }catch(Exception $ex){
            $this->logError($ex->getMessage());
        }
    }

    /**
     * Get changed product from SW-retail in de last x minutes and create a task to synchronize them.
     *
     * @param int $minutes
     */
    public function getChangedSwProducts($minutes = 60){
        $swretail_options = get_option('swretail_options');

        if(!$this->checkLicence()){return;};

        if(!empty($swretail_options['sync']) && $swretail_options['sync'] == 'stock_only'){
            m2m_sw_slack('getChangedSwProducts but setting is stock_only');
            return; 
        }
        
        try{
            swretailUpdateApiRequestCount('article_changed');
            $response = wp_remote_get( $this->swURL.'article_changed/'.$minutes, $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getChangedSwProducts()');
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode)){
                    $responseChunks = array_chunk($response, 10);
                    $timeoffset = 1;
                    foreach ($responseChunks as $chunk){
//						$this->createSwTask('update_or_create_products', implode(',', $chunk));
                        wp_schedule_single_event( time() + (60 * $timeoffset), 'swretail_event_process_changes', [implode(',', $chunk)]);
                        $timeoffset++;
                    }
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getChangedSwProducts()');
                }
            }
        }catch(Exception $ex){
            $this->logError($ex->getMessage());
        }
    }

    /**
     * Get product stock changes from SW-retail in de last x minutes and update stock.
     *
     * @param int $minutes
     */
    public function getChangedSwStock($minutes = 5){
        if(!$this->checkLicence()){return;};
        try{
            swretailUpdateApiRequestCount('article_stock_changed');
            $response = wp_remote_get( $this->swURL.'article_stock_changed/'.$minutes, $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getChangedSwStock(), '. $this->swURL.'article_stock_changed/'.$minutes);
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode)){
                    $responseChunks = array_chunk($response, 10);
                    $timeoffset = 0;
                    foreach ($responseChunks as $chunk){
                        wp_schedule_single_event( time() + (60 * $timeoffset), 'swretail_event_process_changes', [implode(',', $chunk), true]);
                        $timeoffset++;
                    }
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getChangedSwStock()');
                }
            }
        }catch(Exception $ex){
            $this->logError($ex->getMessage());
        }
    }

    /**
     * Check for valid licence.
     *
     * @return boolean true if valid, false on invalid or error
     */
    public function checkLicence(){
        $licence = get_option('swretail_licence');
        // $licence = '';

        $plugin_data = get_plugin_data( dirname(__FILE__).'/swretail.php' );
        $plugin_version = $plugin_data['Version'];

        $connected = true;
        if(empty($licence) || $licence < strtotime('-3 days')
            // && $licence > strtotime("-1 week")
        ){
            $connected = false;
            try{
                $post = [];
                $post['body']['key'] = $this->swLicenceKey;
                $post['body']['domain'] = site_url();
                $post['body']['version'] = $plugin_version;
                $postResponse = wp_remote_post( 'https://swretail.mull2media.nl/licence/', $post );
                if(is_wp_error($postResponse)) {
                    throw new Exception('WP error: #'.$postResponse->get_error_code().' - '.$postResponse->get_error_message().', checkLicence()');
                    $connected = false;
                }else{
                    $postResponse = json_decode($postResponse['body']);
                    if(!isset($postResponse->errorcode)){
                        update_option('swretail_licence', time());
                        $connected = true;
                    }else{
                        throw new Exception('Licence error: #'.$postResponse->errorcode.' - '.$postResponse->message);
                        $connected = false;
                    }
                }
            }catch(Exception $ex){}
        // }elseif($licence < strtotime("-1 week")){
        //     $connected = false;
        //     try{
        //         $post = [];
        //         $post['body']['key'] = $this->swLicenceKey;
        //         $post['body']['domain'] = site_url();
        //         $post['body']['version'] = $plugin_version;
        //         $postResponse = wp_remote_post( 'https://swretail.mull2media.nl/licence/', $post );
        //         if(is_wp_error($postResponse)) {
        //             throw new Exception('WP error: #'.$postResponse->get_error_code().' - '.$postResponse->get_error_message().', checkLicence()');
        //             $connected = false;
        //         }else{
        //             $postResponse = json_decode($postResponse['body']);
        //             if(!isset($postResponse->errorcode)){
        //                 update_option('swretail_licence', time());
        //                 $connected = true;
        //             }else{
        //                 throw new Exception('Licence error: #'.$postResponse->errorcode.' - '.$postResponse->message);
        //                 $connected = false;
        //             }
        //         }
        //     }catch(Exception $ex){
        //         $this->logError($ex->getMessage());
        //         return false;
        //     }
        }else{
            //
        }
        if($connected){ update_option('swretail_licence', time()); }
        return $connected;
    }

    public function checkAddon($addon = ''){

        if(!empty($addon)) {
            try {
                $post = [];
                $post['body']['key'] = $this->swLicenceKey;
                $post['body']['addon'] = $addon;
                $postResponse = wp_remote_post('https://swretail.mull2media.nl/addon/', $post);
                if (is_wp_error($postResponse)) {
                    throw new Exception('WP error: #' . $postResponse->get_error_code() . ' - ' . $postResponse->get_error_message() . ', checkAddon()');
                } else {
                    $postResponse = json_decode($postResponse['body']);

                    if (!isset($postResponse->errorcode)) {
//                    update_option('swretail_licence', time());
                        return true;
                    } else {
                        throw new Exception('Licence error: #' . $postResponse->errorcode . ' - ' . $postResponse->message);
                    }
                }
            } catch (Exception $ex) {
                return false;
            }
        }else{
            return 'Enter addon name';
        }
    }

    public function logError($message){

        $plugin_data = get_plugin_data( dirname(__FILE__).'/swretail.php' );
        $plugin_version = $plugin_data['Version'];
        $filename = 'log-'.str_replace('.','-',$plugin_version);

        if(!file_exists(dirname(__FILE__).'/log/'.$filename.'.txt')){
            fopen(dirname(__FILE__).'/log/'.$filename.'.txt', 'w');
        }
        if(is_writable(dirname(__FILE__).'/log/'.$filename.'.txt')){
            $log = file_get_contents(dirname(__FILE__).'/log/'.$filename.'.txt');
            file_put_contents(dirname(__FILE__).'/log/'.$filename.'.txt', date('Y-m-d H:i:s').' - '.$message."\r\n".$log);
            update_option('swretail_last_error', time());
        }
    }

    

    /**
     * Gets all available warehouses.
     * 
     * @return object $response
     */
    public function getWarehouses(){

        if(!$this->checkLicence()){return;};

        try{
            swretailUpdateApiRequestCount('warehouses');
            $response = wp_remote_get( $this->swURL.'warehouses', $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getWarehouses()');
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode)) {
                    if(gettype($response) != 'array'){
                        $response = [$response];
                    }
                    return $response;
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getWarehouses()');
                }
            }
        }catch(Exception $ex){
            $this->logError($ex);
        }
    }

    
    /**
     * Gets stock from warehouse.
     * 
     * @param string $productsID
     * @param string $warehouseID
     * @param string $position
     * @return object $response
     */
    public function getProductWarehousesStock($productsID = '', $warehouseID = '', $position = '1'){

        if(!$this->checkLicence()){return;};

        if(empty($productsID) || empty($warehouseID)){}

        try{
            swretailUpdateApiRequestCount('article_stock');
            $response = wp_remote_get( $this->swURL.'article_stock/'.$productsID.'/'.$position.'/'.$warehouseID, $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getWarehouses()');
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode)) {
                    if(gettype($response) != 'array'){
                        $response = [$response];
                    }
                    return $response[0]->stock;
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getWarehouses()');
                }
            }
        }catch(Exception $ex){
            $this->logError($ex);
        }
    }

    /**
     * Get product by id and return the data.f
     *
     * @param string $productsID
     * @return object $response
     */
    public function getSwProductById($productsID = ''){
        if(!empty($productsID)){
            try{
                swretailUpdateApiRequestCount('article');
                $response = wp_remote_get( $this->swURL.'article/'.$productsID, $this->wpRemoteArgs );
                //Check for WP errors
                if(is_wp_error($response)) {
                    throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getSwProductById()');
                }else{
                    $response = json_decode($response['body']);
                    //Check for SW-Retail errors
                    if(!isset($response->errorcode)) {
                        return($response);
                    }else{
                        throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getSwProductById()');
                    }
                }
            }catch(Exception $ex){
                $this->logError($ex);
                return $ex;
            }
        }
    }

    /**
     * Get products by ids and run productsUpdateOrCreate() or productsUpdateStock().
     *
     * @param string $productsIDs , separated string of ids
     * @param boolean $stockOnly update stock only
     */
    public function getSwProductsByIds($productsIDs = '', $stockOnly = false){
        global $wpdb;

        if(!$this->checkLicence()){return;};

        if(!empty($productsIDs)){
            try{
                $prodIdsArray = explode(',',$productsIDs);
                swretailUpdateApiRequestCount('article');
                $response = wp_remote_get( $this->swURL.'article/'.$productsIDs, $this->wpRemoteArgs );
                //Check for WP errors
                if(is_wp_error($response)) {
                    throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getSwProductsByIds()');
                }else{
                    $response = json_decode($response['body']);
                    //Check for SW-Retail errors
 

                    if(!isset($response->errorcode)) {
                        if(gettype($response) != 'array'){
                            $response = [$response];
                        }
                        // if($stockOnly){
                        //     $this->productsUpdateStock($response);
                        // }else{
                        //     $this->productsUpdateOrCreate($response);
                        // }
                        $this->productsUpdateOrCreate($response);
                    }else{
                        // $response->errorstring == ID 5000 not found!
                        $exploded_errorstring = explode(' ',$response->errorstring);
                        $removed_sw_prod_id = $exploded_errorstring[1];
                        $meta = $wpdb->get_row("SELECT post_id FROM $wpdb->postmeta WHERE meta_key = '_sw_article_id' AND meta_value = '".$removed_sw_prod_id. "'");
                        if(!is_null($meta)){
                            wp_delete_post($meta->post_id, true);
                            // return $meta->post_id.' deleted';
                            m2m_sw_slack($meta->post_id.' deleted');

                            // check the rest again
                            $prodIdsArray = array_diff($prodIdsArray, array($removed_sw_prod_id));
                            wp_schedule_single_event( time() + 60, 'swretail_event_process_changes', [implode(',',$prodIdsArray)]);
                            
                        }else{
                            throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getSwProductsByIds()');
                        }
                    }
                }
            }catch(Exception $ex){
                $this->logError($ex);
            }
        }
    }

    /**
     * Get all products from SW-Retail than create a batch file for later processing and a corresponding task.
     * This function should only be used for initial import.
     *
     * @param int $page
     * @return array
     */
    public function getAllSwProducts($page = 0,$categories_only = false){
        if(!$this->checkLicence()){return;};
        $swretail_categories = (get_option('swretail_categories') ? get_option('swretail_categories') : array() );
// 		$page = 0;
        try{
            swretailUpdateApiRequestCount('article');
            $response = wp_remote_get( $this->swURL.'article/-1/'.$page, $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getAllSwProducts()');
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode)){
                    if(!empty($response->next_page) && $response->next_page != -1){
                        $next_page = $response->next_page;
                    }else{
                        $next_page = -1;
                    }
                    if(!empty($response->next_page)){
                        unset($response->next_page);
                    }

                    // As from version 2.11.3 don't update products immediately but create task
                    // $updateOrCreateResults = $this->productsUpdateOrCreate($response);
                    $pageProdIds = array();
                    foreach ($response as $swProduct){

                        if( !in_array($swProduct->article_subgroup,$swretail_categories) ){
                            if(!empty($swProduct->article_subgroup)) {
                                $swretail_categories[] = $swProduct->article_subgroup;
                            }
                        }
                        if( !in_array($swProduct->article_group,$swretail_categories) ){
                            if(!empty($swProduct->article_group)) {
                                $swretail_categories[] = $swProduct->article_group;
                            }
                        }

                        $pageProdIds[] = $swProduct->article_id;
                    }
                    // Only execute when clicked op "Get all products" (not categories)
                    if($categories_only === false){
                        wp_schedule_single_event( time() + (60 * ($page+1)), 'swretail_event_process_changes', [implode(',', $pageProdIds)]);
                    }

                    if(!empty($swretail_categories)) {
                        update_option('swretail_categories', $swretail_categories);
                    }

                    $return = ['nextpage' => $next_page];
                    // if($updateOrCreateResults){
                    //     $return['error'] = $updateOrCreateResults;
                    // }
                    return $return;
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getAllSwProducts()');
                }
            }
        }catch(Exception $ex){
            $this->logError($ex);
            return ['error' => $ex];
        }
    }

     /**
     * Check all products in WooCommerce, remove products that are not in SW-Retail anymore
     * Creates batches for checking products
     *
     */
    public function getAllWcProducts(){
        $swretail_addons = get_option('swretail_addons');
        $swretail_options = get_option('swretail_options');

        if(!$this->checkLicence()){return;}
        if(isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1){return;}
        if(isset($swretail_options['delete_products']) && $swretail_options['delete_products'] == 'no'){return;}

        m2m_sw_slack('Check products to delete');

        $offset = 0;
        while($offset == 0 || count($posts) == 10){
            $sw_ids = [];
            $args = array('post_type'=>'product','numberposts' => 10, 'offset'=> $offset, 'fields' => 'ids');
            $posts = get_posts($args);
            foreach($posts as $post_id){
                $sw_ids[] = get_post_meta($post_id,'_sw_article_id',true);
            }                        
            // Schedule event 15 minutes apart to reduce amount of API calls
            wp_schedule_single_event( time() + (15*60 * (($offset+10)/10)), 'swretail_event_process_changes', [implode(',', $sw_ids)]);
            $offset += 10;
        }  

    }

    private function getWcProductIdBySwArticleId($swArticleId){
        global $wpdb;
        $meta = $wpdb->get_row("SELECT post_id FROM $wpdb->postmeta WHERE meta_key = '_sw_article_id' AND meta_value = '" . $swArticleId . "'");
        return (is_null($meta) ? 0 : $meta->post_id);
    }

    private function getWcProductIdBySwBarcode($eans){
        global $wpdb;
        $swretail_options = get_option('swretail_options');

        if(isset($swretail_options['barcode_meta']) && !empty($swretail_options['barcode_meta'])){
            $meta = $wpdb->get_results("SELECT post_id, meta_value FROM $wpdb->postmeta WHERE meta_key = '".$swretail_options['barcode_meta']."' AND meta_value IN('".implode('\',\'', $eans)."') LIMIT 1");
        }else{
			foreach($eans as $ean){
				$meta = $wpdb->get_results("SELECT post_id, meta_value FROM $wpdb->postmeta WHERE meta_key = '_sku' AND meta_value = '$ean' LIMIT 1");
				if(!empty($meta)){
					return $meta[0]->post_id;
				}
			}			
            //$meta = $wpdb->get_results("SELECT post_id, meta_value FROM $wpdb->postmeta WHERE meta_key = '_sku' AND meta_value IN(".implode(',', $eans).") LIMIT 1");
        }
        return (empty($meta) ? null : $meta[0]->post_id);
    }

    private function getWcProductIdsBySwArticleIds($swArticleIds = false){
        global $wpdb;
        if(!empty($swArticleIds)){
            $metas = $wpdb->get_results("SELECT post_id, meta_value FROM $wpdb->postmeta WHERE meta_key = '_sw_article_id' AND meta_value IN(".implode(',', $swArticleIds).")");
        }else{
            $metas = $wpdb->get_results("SELECT post_id, meta_value FROM $wpdb->postmeta WHERE meta_key = '_sw_article_id'");
        }
        $ids = [];
        foreach($metas as $meta){
            $ids[$meta->meta_value] = $meta->post_id;
        }
        return $ids;
    }

    private function getWcVariationIdBySwArticleId($swArticleId, $swVariationPos)
    {
        global $wpdb;
        $meta = $wpdb->get_row("SELECT * FROM $wpdb->postmeta WHERE meta_key = '_sw_variation_id' AND meta_value = '" . $swArticleId . "_" . $swVariationPos . "'");
        if ($post = $wpdb->get_row("SELECT * FROM $wpdb->posts WHERE id = " . $meta->post_id)) {
            if ($post->post_status == 'trash') {
                wp_delete_post($post->ID, true);
                return 0;
            }else{
                return $post->ID;
            }
        } else {
            $wpdb->delete($wpdb->postmeta, array('meta_id' => $meta->meta_id));
            return 0;
        }
    }
    /**
     * Update or create WooCommerce products by SW-Retail products
     *
     * @param object $swProducts
     * @return mixed
     */
    public function productsUpdateOrCreate($swProducts){

        $productIDs = $this->getWcProductIdsBySwArticleIds();
        $productAttributeMapping = get_option('swretail_mapping');
        $swretail_options = get_option('swretail_options');
        $swretail_addons = get_option('swretail_addons');
        $swretail_meta = get_option('swretail_meta');
        $swretail_categories = (get_option('swretail_categories') ? get_option('swretail_categories') : array() );
        $swretail_combine_attributes = (get_option('swretail_combine_attributes') ? get_option('swretail_combine_attributes') : array() );


        //Allow backorders?
        $backorders = 'no';
        if(!empty($swretail_options['backorders'])){
            $backorders = $swretail_options['backorders'];
            if($swretail_options['backorders'] == 'false' || $swretail_options['backorders'] == 'false-notify'){
                $backorders = false;
            }
        }else{
            $backorders = 'no';
        }
        
        $errors = [];
        foreach ($swProducts as $swProduct){
            
            
            try{

                if( !in_array($swProduct->article_subgroup,$swretail_categories) ){
                    if(!empty($swProduct->article_subgroup)) {
                        $swretail_categories[] = $swProduct->article_subgroup;
                    }
                }
                if( !in_array($swProduct->article_group,$swretail_categories) ){
                    if(!empty($swProduct->article_group)) {
                        $swretail_categories[] = $swProduct->article_group;
                    }
                }

                switch($swretail_options['sku_field']){
                    case 'article_id_artfabr':
                        $swProduct->article_sku = $swProduct->article_id.(!empty($swProduct->article_artfabr) ? ' | '.$swProduct->article_artfabr : '');
                        break;
                    case 'article_numb':
                        $swProduct->article_sku = $swProduct->article_number;
                        break;
                    case 'barcode_13':
                        foreach($swProduct->barcodes as $barcode){
                            if((strlen($barcode->barcode) == 13) || isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1){
                                $swProduct->article_sku = $barcode->barcode;
                            }
                        }
                        break;
                    default:
                        $swProduct->article_sku = $swProduct->{$swretail_options['sku_field']};
                }

                $productID = $this->getWcProductIdBySwArticleId($swProduct->article_id);

                // Check if products needs to be in this webshop
                $inWebshop = false;
                if(isset($swProduct->article_inwebshop) && $swProduct->article_inwebshop == 1){
                    $inWebshop = true;
                    if( isset($swretail_options['category']) && !empty($swretail_options['category']) ){
                        $inWebshop = false;
                        
                        if(in_array(sanitize_title($swProduct->article_subgroup),array_keys($swretail_options['category'])) || in_array(sanitize_title($swProduct->article_group),array_keys($swretail_options['category']))){
                            $inWebshop = true;
                        }
                        foreach($swProduct->article_group_extra[0] as $article_group_extra){
                            if(in_array(sanitize_title($article_group_extra),array_keys($swretail_options['category']))){
                                $inWebshop = true;
                            }
                        }
                    }
                }elseif(isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1) {
                    $eans = [];
                    if(!empty($swProduct->barcodes)){
                        foreach ($swProduct->barcodes as $barcode){
                            $eans[] = $barcode->barcode;
                        }
                    }
                    if($productID = $this->getWcProductIdBySwBarcode($eans)){
                        $inWebshop = true;
                    }
                }
                if($inWebshop){

                    if(isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1) {
                        $eans = [];
                        if(!empty($swProduct->barcodes)){
                            foreach ($swProduct->barcodes as $barcode){
                                $eans[] = $barcode->barcode;
                            }
                        }
                        $productID = $this->getWcProductIdBySwBarcode($eans);
                    }
//                    else{
//                        $productID = $this->getWcProductIdBySwArticleId($swProduct->article_id);
//                    }

                    //If variable

                    if(!empty($swProduct->sizes) && isset($swProduct->sizes[0]->description) && $swProduct->sizes[0]->description != ''){
                        $product = new WC_Product_Variable($productID);
                        if(!isset($swretail_options['no_sync']['title']) || $swretail_options['no_sync']['title'] != 1){
                            $product->set_name($swProduct->article_description);
                        }
                        $product->set_sku($swProduct->article_sku);


                        if(!isset($swretail_options['no_sync']['price']) || $swretail_options['no_sync']['price'] != 1){ 
                            if (isset($swProduct->article_price_web) && !empty($swProduct->article_price_web) && (int)$swProduct->article_price_web != 0 && $swretail_options['webprice'] != 'discount') {
                                $regularPrice = str_replace(',','.',$swProduct->article_price_web);
                            } else {
                                $regularPrice = str_replace(',','.',$swProduct->article_baseprice);
                            }

                            // // ADDON: Foreign currency calculator
                            // if(isset($swretail_addons['currency_calculator']['active']) && $swretail_addons['currency_calculator']['active'] == 1){
                            //     if(isset($swretail_addons['currency_calculator']['currency_to']) && isset($swretail_addons['currency_calculator']['currencies'][$swretail_addons['currency_calculator']['currency_to']])){
                            //         $oldCurrencyPrice = (int)$regularPrice;
                            //         $newCurrencyPrice = ($oldCurrencyPrice*$swretail_addons['currency_calculator']['currencies'][$swretail_addons['currency_calculator']['currency_to']])

                            //         if(isset($swretail_addons['currency_calculator']['round_prices']) && $swretail_addons['currency_calculator']['round_prices'] == 'up_ending_5_10'){
                            //             $newCurrencyPrice = round($newCurrencyPrice/5)*5;
                            //         }

                            //         slack($oldCurrencyPrice.' '.$swretail_addons['currency_calculator']['currency_from'].' = '.$newCurrencyPrice.' '.$swretail_addons['currency_calculator']['currency_to']);
                            //     }
                            // }
                            // $regularPrice = sw_calculate_price($regularPrice);
                            $product->set_regular_price(sw_calculate_price($regularPrice,$swProduct->article_taxrate));

                            if(!isset($swretail_options['no_sync']['sale_price']) || $swretail_options['no_sync']['sale_price'] != 1){ 
                                if( $swretail_options['webprice'] == 'discount' && isset($swProduct->article_price_web) && !empty($swProduct->article_price_web) && (int)$swProduct->article_price_web != 0 ){
                                    $articlePriceWeb = sw_calculate_price(str_replace(',','.',$swProduct->article_price_web),$swProduct->article_taxrate);
                                    $product->set_sale_price( str_replace(',','.',$swProduct->article_price_web) );
                                }elseif (isset($swProduct->article_price_web_discount) && !empty($swProduct->article_price_web_discount)) {
                                    $salePrice = $this->getDiscount($regularPrice, str_replace(',','.',$swProduct->article_price_web_discount) );
                                    $salePrice = sw_calculate_price($salePrice,$swProduct->article_taxrate);
                                    $product->set_sale_price($salePrice);
                                    $product->set_price($salePrice);
                                } elseif (isset($swProduct->article_discount) && !empty($swProduct->article_discount) ) {
                                    $salePrice = $this->getDiscount($regularPrice, str_replace(',','.',$swProduct->article_discount));
                                    $salePrice = sw_calculate_price($salePrice,$swProduct->article_taxrate);
                                    $product->set_sale_price($salePrice);
                                    $product->set_price($salePrice);
                                } else {
                                    $product->set_sale_price('');
                                    $product->set_price($regularPrice);
                                }
                            }

                            //Wholesale pricing
                            if (isset($swretail_addons['wwpp_wholesale']['active']) && $swretail_addons['wwpp_wholesale']['active'] == 1) {
                                if (isset($swProduct->article_price_wholesale) && !empty($swProduct->article_price_wholesale) && $swProduct->article_price_wholesale != 0) {
                                    $wholesalePrice = $swProduct->article_price_wholesale * (($swProduct->article_taxrate / 100) + 1);
                                    $wholesalePrice = sw_calculate_price($wholesalePrice,$swProduct->article_taxrate);
                                    $product->update_meta_data('wholesale_customer_wholesale_price', $wholesalePrice);
                                    $product->update_meta_data('wholesale_customer_have_wholesale_price', 'yes');
                                    // $product->update_meta_data('wwpp_product_wholesale_visibility_filter', 'all'); 
                                } else {
                                    $product->update_meta_data('wholesale_customer_wholesale_price', '');
                                    $product->update_meta_data('wholesale_customer_have_wholesale_price', 'no');
                                    $product->update_meta_data('wwpp_product_wholesale_visibility_filter', 'all');
                                }
                            }

                            if ($swProduct->article_taxrate == 9.00 && !empty($swretail_options['tax_rate_reduced'])) {
                                $product->set_tax_class($swretail_options['tax_rate_reduced']);
                            } elseif (!empty($swretail_options['tax_rate_default'])) {
                                $product->set_tax_class($swretail_options['tax_rate_default']);
                            } else {
                                $product->set_tax_class('standard');
                            }
                            $product->set_manage_stock(0); // 0 = off, 1 = on
                        }

                        //Backorders
                        if($backorders) {
                            $product->set_backorders($backorders); //Options: 'yes', 'no' or 'notify'.
                        }else{
                            if(isset($swProduct->article_run_out)){
                                if($swProduct->article_run_out == 1){
                                    // if($backorders == 'yes') {
                                    //     $product->set_backorders('yes');
                                    // }elseif($backorders == 'no'){
                                        $product->set_backorders('no');
                                    // }
                                }else{
                                    if($swretail_options['backorders'] == 'false-notify'){
                                        $product->set_backorders('notify');
                                    }else{
                                        $product->set_backorders('yes');
                                    }
                                }
                            }
                        }


                        if(!isset($swretail_options['no_sync']['description']) || $swretail_options['no_sync']['description'] != 1){
                            if (isset($swProduct->article_additional_info) && !empty($swProduct->article_additional_info)) {
                                $product->set_description($swProduct->article_additional_info);
                            }else{
                                $product->set_description('');
                            }
                            if(isset($swProduct->article_memo) && !empty($swProduct->article_memo)) {
                                $product->set_short_description($swProduct->article_memo);
                            }else{
                                $product->set_short_description('');
                            }
                        }
                        
                        if(!empty($swProduct->article_homepage) && $swProduct->article_homepage == 1) {
                            $product->set_featured(true);
                        }else{
                            $product->set_featured(false);
                        }

                        $product->update_meta_data('_sw_article_id', $swProduct->article_id);
                        if(!isset($swretail_options['no_sync']['meta']) || $swretail_options['no_sync']['meta'] != 1){
                            if(is_plugin_active('wordpress-seo/wp-seo.php') || is_plugin_active('wordpress-seo-premium/wp-seo-premium.php')){
                                $product->update_meta_data('_yoast_wpseo_title', $swProduct->article_metatitle);
                                if(!empty($swProduct->article_metakeywords)){
                                    $product->update_meta_data('_yoast_wpseo_focuskw', $swProduct->article_metakeywords);
                                }
                                $product->update_meta_data('_yoast_wpseo_metadesc', $swProduct->article_metadescription);
                            }elseif(is_plugin_active('seo-by-rank-math/rank-math.php') || is_plugin_active('seo-by-rank-math-pro/rank-math-pro.php')){
                                $product->update_meta_data('rank_math_title', $swProduct->article_metatitle);
                                if(!empty($swProduct->article_metakeywords)){
                                    $product->update_meta_data('rank_math_focus_keyword', $swProduct->article_metakeywords);
                                }
                                $product->update_meta_data('rank_math_description', $swProduct->article_metadescription);
                            }
                        }

                        // Crossell
                        $crossell = [];
                        if(!empty($swProduct->article_crossell)){
                            foreach($swProduct->article_crossell as $a){
                                (!empty($productIDs[$a]) ? $crossell[] = $productIDs[$a] : '');
                            }
                        }
                        $product->set_cross_sell_ids($crossell);

                        // Upsell
                        $upsell = [];
                        if(!empty($swProduct->article_upsell)){
                            foreach($swProduct->article_upsell as $a){
                                (!empty($productIDs[$a]) ? $upsell[] = $productIDs[$a] : '');
                            }
                        }
                        $product->set_upsell_ids($upsell);

                        /**
                         * @var int number of variations
                         */
                        $variationCount = 0;
                        /**
                         * @var array sizes
                         */
                        $productSizeList = [];
                        /**
                         * @var array of product variation objects (name, stock, sku)
                         */
                        $productVariations = [];
                        foreach ($swProduct->sizes as $productSize){
                            $productVariations[$productSize->position] = new stdClass();
                            $productVariations[$productSize->position]->name = str_replace('  ', ' ', (string)$productSize->description);
                            if(!isset($swretail_options['no_sync']['stock']) || $swretail_options['no_sync']['stock'] != 1) {
                                $productVariations[$productSize->position]->stock = $productSize->stock;

                                if(isset($swretail_options['stock_location']) && !empty($swretail_options['stock_location'])){
                                    $productVariations[$productSize->position]->stock = $this->getProductWarehousesStock($swProduct->article_id,$swretail_options['stock_location'],$productSize->position);
                                }

                            }
                            $productVariations[$productSize->position]->salepricedelta = $productSize->salepricedelta;
                            $productSizeList[] = str_replace('  ', ' ', (string)$productSize->description);
                            $variationCount++;
                        }

                        //continue to next product if variable product doesn't have any variations
                        if($variationCount == 0){
                            break;
                        }

                        if(!isset($swretail_options['no_sync']['attributes']) || $swretail_options['no_sync']['attributes'] != 1){

                            if(!empty($swretail_combine_attributes)){
                                // do mapping
                            }
                            
                            $productAttributes = [];
                            foreach ($productAttributeMapping as $attribute_key => $attribute_taxonomy) {
                                if (empty($attribute_taxonomy)) {
                                    continue;
                                }
                                if($attribute_taxonomy == '_length'){
                                    $product->set_length( $swProduct->fields->{$attribute_key} );
                                }
                                if($attribute_taxonomy == '_width'){
                                    $product->set_width( $swProduct->fields->{$attribute_key} );
                                }
                                if($attribute_taxonomy == '_height'){
                                    $product->set_height( $swProduct->fields->{$attribute_key} );
                                }
                                $attribute_taxonomy = explode(':', $attribute_taxonomy);
                                $options = [];
                                $isVariation = 0;
                                switch ($attribute_key) {
                                    case 'size':
                                        $options = $productSizeList;
                                        $isVariation = 1;
                                        $sizeAtribute = $attribute_taxonomy[1];
                                        break;
                                    case 'color':
                                        if(isset($swretail_combine_attributes[$attribute_taxonomy[1]]) && !empty($swProduct->article_color)){
                                            foreach($swretail_combine_attributes[$attribute_taxonomy[1]] as $key => $value){
                                                if(strpos($value,$swProduct->article_color) !== false){
                                                    $options = [$key];
                                                }else{
                                                    $options = [$swProduct->article_color];
                                                }
                                            }
                                        }else{
                                            $options = [$swProduct->article_color];
                                        }
                                        break;
                                    case 'freefield1':
                                        $options = [$swProduct->article_freefield1];
                                        break;
                                    case 'freefield2':
                                        $options = [$swProduct->article_freefield2];
                                        break;
                                    case 'brand':
                                        $options = [$swProduct->article_manufacturer];
                                        break;
                                    default:
                                        if (!empty($swProduct->fields->{$attribute_key})) {
                                            
                                            if(isset($swretail_combine_attributes[$attribute_taxonomy[1]])){
                                                $option_found = false;
                                                foreach($swretail_combine_attributes[$attribute_taxonomy[1]] as $key => $value){
                                                    if(strpos($value,$swProduct->fields->{$attribute_key}) !== false){
                                                        $term = get_term_by('slug', $key, $attribute_taxonomy[1]);
                                                        if($term){
                                                            $options = [$term->name];
                                                            $option_found = true;
                                                        }
                                                    }
                                                    if(!$option_found){
                                                        $options = explode('', $swProduct->fields->{$attribute_key});
                                                    }
                                                }
                                            }else{
                                                $options = explode('', $swProduct->fields->{$attribute_key});
                                            }

                                        }
                                }
                                if (!empty($options)) {
                                    $attribute = new WC_Product_Attribute();
                                    $attribute->set_id($attribute_taxonomy[0]);
                                    $attribute->set_name($attribute_taxonomy[1]);
                                    $attribute->set_visible(1);
                                    $attribute->set_variation($isVariation);
                                    $attribute->set_options($options);
                                    $productAttributes[] = $attribute;
                                }
                            }

                            // Set attributes
                            $product->set_attributes($productAttributes);
                            $productAttributes = $product->get_attributes();
                        }

                        if(!isset($swretail_options['no_sync']['categories']) || $swretail_options['no_sync']['categories'] != 1){
                            // Set categories
                            $categoryIds = [];
                            if(!empty($swProduct->article_group)){
                                $category = $this->categoryFindOrCreate($swProduct->article_group);
                                $categoryIds[] = $category;
                                if(!empty($swProduct->article_subgroup)){
                                    $subCategory = $this->categoryFindOrCreate($swProduct->article_subgroup, $category);
                                    $categoryIds[] = $subCategory;
                                    if (!empty($swProduct->article_subsubgroup)) {
                                        $subSubCategory = $this->categoryFindOrCreate($swProduct->article_subsubgroup, $subCategory);
                                        $categoryIds[] = $subSubCategory;
                                    }
                                }
                            }
                            // Extra category groups
                            if(!empty($swProduct->article_group_extra)){
                                foreach($swProduct->article_group_extra as $extra_categories){
                                    if(!empty($extra_categories[0])){
                                        $category = $this->categoryFindOrCreate($extra_categories[0]);
                                        $categoryIds[] = $category;
                                        if(!empty($extra_categories[1])){
                                            $subCategory = $this->categoryFindOrCreate($extra_categories[1], $category);
                                            $categoryIds[] = $subCategory;
                                            if (!empty($extra_categories[2])) {
                                                $subSubCategory = $this->categoryFindOrCreate($extra_categories[2], $subCategory);
                                                $categoryIds[] = $subSubCategory;
                                            }
                                        }
                                    }
                                }
                            }

                            // 3.7.0: Add product with a discount price to a specific SALE category
                            if(  isset($swretail_addons['auto_sale_cat']['active']) && $swretail_addons['auto_sale_cat']['active'] == 1 && isset($swretail_options['auto_sale_cat']) ){
                                if($swProduct->article_discount){
                                    $categoryIds[] = $swretail_options['auto_sale_cat'];
                                }elseif($swretail_options['webprice'] == 'discount' && $swProduct->article_price_web && $swProduct->article_price_web < $swProduct->article_baseprice){
                                    $categoryIds[] = $swretail_options['auto_sale_cat'];
                                }
                            }

                            if(!empty($categoryIds)){
                                $product->set_category_ids(array_unique($categoryIds));
                            }
                        }

                        $product->save();

                        //Sync custom meta fields
                        if(!empty($swretail_meta)){
                            foreach($swretail_meta as $meta){
                                if($meta['value'] == 'barcode'){
                                    $product->update_meta_data($meta['key'], $swProduct->barcodes[0]->barcode);
                                }elseif($meta['value'] == 'barcode2'){
                                    $product->update_meta_data($meta['key'], $swProduct->barcodes[1]->barcode);
                                }else{
                                    if(isset($swProduct->{$meta['value']}) && !empty($swProduct->{$meta['value']})){
                                        $product->update_meta_data($meta['key'], $swProduct->{$meta['value']});
                                    }elseif(isset($swProduct->fields->{$meta['value']}) && !empty($swProduct->fields->{$meta['value']})){
                                        $product->update_meta_data($meta['key'], $swProduct->fields->{$meta['value']});
                                    }else{
                                        delete_post_meta($product->get_ID(),$meta['key']);
                                    }
                                }
                            }
                        }

                        //Get product size attribute name slug pairs for variation mapping.
                        $productSizes = $productAttributes[$sizeAtribute]->get_terms();
                        $tmpProductSizes = [];
                        foreach($productSizes as $productSize){
                            $tmpProductSizes[sanitize_title($productSize->name)] = $productSize->slug;
                        }
                        $productSizes = $tmpProductSizes;

                        //Update or create variations
                        $variationIdList = [];
                        foreach ($productVariations as $productVariationPosition => $productVariation){
                            $variationID = $this->getWcVariationIdBySwArticleId($swProduct->article_id, $productVariationPosition);
                            $variation = new WC_Product_Variation($variationID);
                            $variation->update_meta_data('_sw_size_pos', $productVariationPosition);
                            $variation->update_meta_data('_sw_variation_id', $swProduct->article_id.'_'.$productVariationPosition);

                            if(!isset($swretail_options['no_sync']['attributes']) || $swretail_options['no_sync']['attributes'] != 1){
                                // Set weight
                                if(isset($swProduct->article_weight)) {
                                    $variation->set_weight($swProduct->article_weight);
                                }
                                
                            }

                            if(!isset($swretail_options['no_sync']['price']) || $swretail_options['no_sync']['price'] != 1){
                                $additionalVariationPrice = 0;
                                if (isset($productVariation->salepricedelta) && !empty($productVariation->salepricedelta)) {
                                    $additionalVariationPrice = $productVariation->salepricedelta;
                                }

                                if ($swretail_options['webprice'] != 'discount' && isset($swProduct->article_price_web) && !empty($swProduct->article_price_web) && (int)$swProduct->article_price_web != 0) {
                                    $regularPrice = str_replace(',','.',$swProduct->article_price_web);
                                } else {
                                    $regularPrice = str_replace(',','.',$swProduct->article_baseprice);
                                }

                                // $regularPrice = sw_calculate_price($regularPrice);
                                // $additionalVariationPrice = sw_calculate_price($additionalVariationPrice);

                                $variation->set_regular_price( sw_calculate_price(number_format($regularPrice, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''),$swProduct->article_taxrate) );

                                if(!isset($swretail_options['no_sync']['sale_price']) || $swretail_options['no_sync']['sale_price'] != 1){ 
                                    
                                    if ($swretail_options['webprice'] == 'discount' && isset($swProduct->article_price_web) && !empty($swProduct->article_price_web)) {
                                        // $salePrice = $this->getDiscount($regularPrice, str_replace(',','.',$swProduct->article_price_web_discount) );
                                        // $articlePriceWeb = sw_calculate_price($swProduct->article_price_web);
                                        $variation->set_sale_price( sw_calculate_price(number_format($swProduct->article_price_web, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''),$swProduct->article_taxrate));
                                        $variation->set_price( sw_calculate_price(number_format($regularPrice, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''),$swProduct->article_taxrate));
                                    }elseif (isset($swProduct->article_price_web_discount) && !empty($swProduct->article_price_web_discount)) {
                                        $salePrice = $this->getDiscount($regularPrice, str_replace(',','.',$swProduct->article_price_web_discount) );
                                        $salePrice = sw_calculate_price($salePrice,$swProduct->article_taxrate);
                                        $variation->set_sale_price(number_format($salePrice, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''));
                                        $variation->set_price(number_format($salePrice, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''));
                                    } elseif (isset($swProduct->article_discount) && !empty($swProduct->article_discount) ) {
                                        $salePrice = $this->getDiscount($regularPrice, str_replace(',','.',$swProduct->article_discount));
                                        $salePrice = sw_calculate_price($salePrice,$swProduct->article_taxrate);
                                        $variation->set_sale_price(number_format($salePrice, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''));
                                        $variation->set_price(number_format($salePrice, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''));
                                    } else {
                                        $variation->set_sale_price('');
                                        $variation->set_price( sw_calculate_price(number_format($regularPrice, 2, '.', '') + number_format($additionalVariationPrice, 2, '.', ''),$swProduct->article_taxrate));
                                    }
                                }

                                //Wholesale pricing
                                if (isset($swretail_addons['wwpp_wholesale']['active']) && $swretail_addons['wwpp_wholesale']['active'] == 1) {
                                    if (isset($swProduct->article_price_wholesale) && !empty($swProduct->article_price_wholesale) && $swProduct->article_price_wholesale != 0) {
                                        $wholesalePrice = $swProduct->article_price_wholesale * (($swProduct->article_taxrate / 100) + 1);
                                        $wholesalePrice = sw_calculate_price($wholesalePrice,$swProduct->article_taxrate);
                                        $variation->update_meta_data('wholesale_customer_wholesale_price', $wholesalePrice);
                                        $variation->update_meta_data('wholesale_customer_have_wholesale_price', 'yes');
                                        $variation->update_meta_data('wwpp_product_wholesale_visibility_filter', 'all');
                                    } else {
                                        $variation->update_meta_data('wholesale_customer_wholesale_price', '');
                                        $variation->update_meta_data('wholesale_customer_have_wholesale_price', 'no');
                                        $variation->update_meta_data('wwpp_product_wholesale_visibility_filter', 'all');
                                    }
                                }

                                if ($swProduct->article_taxrate == 9.00 && !empty($swretail_options['tax_rate_reduced'])) {
                                    $variation->set_tax_class($swretail_options['tax_rate_reduced']);
                                } elseif (!empty($swretail_options['tax_rate_default'])) {
                                    $variation->set_tax_class($swretail_options['tax_rate_default']);
                                } else {
                                    $variation->set_tax_class('standard');
                                }

                                //Add deposit price when addon is active
                                if(isset($swretail_addons['deposit']['active']) && $swretail_addons['deposit']['active'] == 1){
                                    if (isset($swProduct->article_deposit) && !empty($swProduct->article_deposit) && $swProduct->article_deposit != 0) {          
                                        $variation->update_meta_data('_swretail_deposit_price', $swProduct->article_deposit);
                                    }else{ 
                                        $variation->delete_meta_data('_swretail_deposit_price');
                                    }
                                }

                            }



                            $variation->set_parent_id($product->get_id());
                            $variation->set_attributes([$sizeAtribute => $productSizes[sanitize_title($productVariation->name)]]);

                            if(!isset($swretail_options['no_sync']['stock']) || $swretail_options['no_sync']['stock'] != 1){
                                $variation->set_manage_stock(1); // 0 = off, 1 = on
                                if($backorders){
                                    $variation->set_backorders($backorders); //Options: 'yes', 'no' or 'notify'.
                                }else{
                                    if(isset($swProduct->article_run_out)){
                                        if($swProduct->article_run_out == 1){
                                                $variation->set_backorders('no');
                                        }else{
                                            $variation->set_backorders('yes');
                                        }
                                    }
                                }
                                $variation->set_stock_quantity($productVariation->stock);

                                // UPDATE 3.9.0: Addon External Stock
                                if(isset($swretail_addons['external_stock']['active']) && $swretail_addons['external_stock']['active'] == 1){
                                    if(get_post_meta($variationID,'_sw_external_stock',true) && get_post_meta($variationID,'_sw_external_stock',true) != 0){
                                        // Only update with external stock when stock has been updated in last 24 hours. Otherwise delete post meta.
                                        if(get_post_meta($variationID,'_sw_external_stock_updated',true) && get_post_meta($variationID,'_sw_external_stock_updated',true) > (current_time('timestamp')-(3600*24))){
                                            $external_stock = get_post_meta($variationID,'_sw_external_stock',true);
                                            $stock = $swProduct->sizes[0]->stock + $external_stock;
                                            $product->set_stock_quantity($stock);
                                        }else{
                                            delete_post_meta($variationID,'_sw_external_stock');
                                            delete_post_meta($variationID,'_sw_external_stock_updated');
                                        }
                                    }
                                }

                            }
                            $variation->save();

                            $variationIdList[] = $variation->get_id();
                        }
                        WC_Product_Variable::sync($product->get_id());

                        //Delete unused variations
                        $variations = $product->get_children();
                        foreach ($variations as $variation){
                            if(!in_array($variation, $variationIdList)){
                                wp_delete_post( $variation, true );
                            }
                        }
                        WC_Product_Variable::sync($product->get_id());

                        if(!isset($swretail_options['no_sync']['images']) || $swretail_options['no_sync']['images'] != 1){
                            //Find or create images
                            $images = $this->imagesFindOrCreate($swProduct->images, $product->get_id());
                            $product->set_image_id($images->mainImage);
                            $product->set_gallery_image_ids($images->galleryImages);
                        }

                        $product->save();
                    }else{ // Simple product

                        $product = new WC_Product($productID);
                        if(!isset($swretail_options['no_sync']['title']) || $swretail_options['no_sync']['title'] != 1){
                            $product->set_name($swProduct->article_description);
                        }


                        if(!isset($swretail_options['no_sync']['price']) || $swretail_options['no_sync']['price'] != 1){
                            if (isset($swProduct->article_price_web) && !empty($swProduct->article_price_web) && $swProduct->article_price_web != 0 && $swretail_options['webprice'] != 'discount') {
                                $regularPrice = str_replace(',','.',$swProduct->article_price_web);
                            } else {
                                $regularPrice = $swProduct->article_baseprice;
                            }
                            // $regularPrice = sw_calculate_price($regularPrice);
                            $product->set_regular_price(sw_calculate_price($regularPrice,$swProduct->article_taxrate));


                            if(!isset($swretail_options['no_sync']['sale_price']) || $swretail_options['no_sync']['sale_price'] != 1){ 
                                if( $swretail_options['webprice'] == 'discount' && isset($swProduct->article_price_web) && !empty($swProduct->article_price_web) && (int)$swProduct->article_price_web != 0 ){
                                    $articlePriceWeb = sw_calculate_price($swProduct->article_price_web,$swProduct->article_taxrate);
                                    $product->set_sale_price(str_replace(',','.',$articlePriceWeb));
                                }elseif (isset($swProduct->article_price_web_discount) && !empty($swProduct->article_price_web_discount)) {
                                    $salePrice = $this->getDiscount($regularPrice, str_replace(',','.',$swProduct->article_price_web_discount));
                                    $salePrice = sw_calculate_price($salePrice,$swProduct->article_taxrate);
                                    $product->set_sale_price($salePrice);
                                    $product->set_price($salePrice);
                                // } elseif (isset($swProduct->article_discount) && !empty($swProduct->article_discount) && (isset($swProduct->article_price_web_discount) && (int)$swProduct->article_price_web_discount != 0)) {
                                } elseif (isset($swProduct->article_discount) && !empty($swProduct->article_discount) ) {
                                    $salePrice = $this->getDiscount($regularPrice, $swProduct->article_discount);
                                    $salePrice = sw_calculate_price($salePrice,$swProduct->article_taxrate);
                                    $product->set_sale_price($salePrice);
                                    $product->set_price($salePrice);
                                } else {
                                    $product->set_sale_price('');
                                    $product->set_price(sw_calculate_price($regularPrice,$swProduct->article_taxrate));
                                }
                            }

                            //Wholesale pricing
                            if (isset($swretail_addons['wwpp_wholesale']['active']) && $swretail_addons['wwpp_wholesale']['active'] == 1) {
                                if (isset($swProduct->article_price_wholesale) && !empty($swProduct->article_price_wholesale) && $swProduct->article_price_wholesale != 0) {
                                    $wholesalePrice = $swProduct->article_price_wholesale * (($swProduct->article_taxrate / 100) + 1);
                                    $wholesalePrice = sw_calculate_price($wholesalePrice,$swProduct->article_taxrate);
                                    $product->update_meta_data('wholesale_customer_wholesale_price', $wholesalePrice);
                                    $product->update_meta_data('wholesale_customer_have_wholesale_price', 'yes');
                                    //$product->update_meta_data('wwpp_product_wholesale_visibility_filter', 'all');
                                } else {
                                    $product->update_meta_data('wholesale_customer_wholesale_price', '');
                                    $product->update_meta_data('wholesale_customer_have_wholesale_price', 'no');
                                    $product->update_meta_data('wwpp_product_wholesale_visibility_filter', 'all');
                                }
                            }

                            if ($swProduct->article_taxrate <= 9.00 && !empty($swretail_options['tax_rate_reduced'])) {
                                $product->set_tax_class($swretail_options['tax_rate_reduced']);
                            } elseif (!empty($swretail_options['tax_rate_default'])) {
                                $product->set_tax_class($swretail_options['tax_rate_default']);
                            } else {
                                $product->set_tax_class('standard');
                            }

                            //Add deposit price when addon is active
                            if(isset($swretail_addons['deposit']['active']) && $swretail_addons['deposit']['active'] == 1){
                                if (isset($swProduct->article_deposit) && !empty($swProduct->article_deposit) && $swProduct->article_deposit != 0) {          
                                    $product->update_meta_data('_swretail_deposit_price', $swProduct->article_deposit);
                                }else{ 
                                    $product->delete_meta_data('_swretail_deposit_price');
                                }
                            }
                        }
                        $product->set_sku($swProduct->article_sku);


                        //Sync custom meta fields
                        if(!empty($swretail_meta)){
                            foreach($swretail_meta as $meta){
                                if($meta['value'] == 'barcode'){
                                    $product->update_meta_data($meta['key'], $swProduct->barcodes[0]->barcode);
                                }elseif($meta['value'] == 'barcode2'){
                                    $product->update_meta_data($meta['key'], $swProduct->barcodes[1]->barcode);
                                }else{
                                    if(isset($swProduct->{$meta['value']}) && !empty($swProduct->{$meta['value']})){
                                        $product->update_meta_data($meta['key'], $swProduct->{$meta['value']});
                                    }elseif(isset($swProduct->fields->{$meta['value']}) && !empty($swProduct->fields->{$meta['value']})){
                                        $product->update_meta_data($meta['key'], $swProduct->fields->{$meta['value']});
                                    }else{
                                        delete_post_meta($product->get_ID(),$meta['key']);
                                    }
                                }
                            }
                        }

                        if(!isset($swretail_options['no_sync']['stock']) || $swretail_options['no_sync']['stock'] != 1){
                            $product->set_manage_stock(1); // 0 = off, 1 = on
                            if($backorders) {
                                $product->set_backorders($backorders); //Options: 'yes', 'no' or 'notify'.
                            }else{
                                if(isset($swProduct->article_run_out)){
                                    if($swProduct->article_run_out == 1){
                                        // if($backorders == 'yes') {
                                        //     $product->set_backorders('yes');
                                        // }elseif($backorders == 'no'){
                                            $product->set_backorders('no');
                                        // }
                                    }else{ 
                                        if($swretail_options['backorders'] == 'false-notify'){
                                            $product->set_backorders('notify');
                                        }else{
                                            $product->set_backorders('yes');
                                        }
                                    }
                                }
                            }
                            $product->set_stock_quantity($swProduct->sizes[0]->stock);
                            // UPDATE 3.9.0: Addon External Stock
                            if(isset($swretail_addons['external_stock']['active']) && $swretail_addons['external_stock']['active'] == 1){
                                if(get_post_meta($productID,'_sw_external_stock',true) && get_post_meta($productID,'_sw_external_stock',true) != 0){
                                    // Only update with external stock when stock has been updated in last 24 hours. Otherwise delete post meta.
                                    if(get_post_meta($productID,'_sw_external_stock_updated',true) && get_post_meta($productID,'_sw_external_stock_updated',true) > (current_time('timestamp')-(3600*24))){
                                        $external_stock = get_post_meta($productID,'_sw_external_stock',true);
                                        $stock = $swProduct->sizes[0]->stock + $external_stock;
                                        $product->set_stock_quantity($stock);
                                    }else{
                                        delete_post_meta($productID,'_sw_external_stock');
                                        delete_post_meta($productID,'_sw_external_stock_updated');
                                    }
                                }
                            }

                            if(isset($swretail_options['stock_location']) && !empty($swretail_options['stock_location'])){
                                $product->set_stock_quantity($this->getProductWarehousesStock($swProduct->article_id,$swretail_options['stock_location']));
                            } 

                        }

                        if(!isset($swretail_options['no_sync']['description']) || $swretail_options['no_sync']['description'] != 1){
                            if (isset($swProduct->article_additional_info) && !empty($swProduct->article_additional_info)) {
                                $product->set_description($swProduct->article_additional_info);
                            }else{
                                $product->set_description('');
                            }
                            if(isset($swProduct->article_memo) && !empty($swProduct->article_memo)){
                                $product->set_short_description($swProduct->article_memo);
                            }else{
                                $product->set_short_description('');
                            }
                        }

                        
                        if(!empty($swProduct->article_homepage) && $swProduct->article_homepage == 1) {
                            $product->set_featured(true);
                        }else{
                            $product->set_featured(false);
                        }
                        $product->update_meta_data('_sw_article_id', $swProduct->article_id);
                        if(!isset($swretail_options['no_sync']['meta']) || $swretail_options['no_sync']['meta'] != 1){
                            if(is_plugin_active('wordpress-seo/wp-seo.php') || is_plugin_active('wordpress-seo-premium/wp-seo-premium.php')){
                                $product->update_meta_data('_yoast_wpseo_title', $swProduct->article_metatitle);
                                if(!empty($swProduct->article_metakeywords)){
                                    $product->update_meta_data('_yoast_wpseo_focuskw', $swProduct->article_metakeywords);
                                }
                                $product->update_meta_data('_yoast_wpseo_metadesc', $swProduct->article_metadescription);
                            }elseif(is_plugin_active('seo-by-rank-math/rank-math.php') || is_plugin_active('seo-by-rank-math-pro/rank-math-pro.php')){
                                $product->update_meta_data('rank_math_title', $swProduct->article_metatitle);
                                if(!empty($swProduct->article_metakeywords)){
                                    $product->update_meta_data('rank_math_focus_keyword', $swProduct->article_metakeywords);
                                }
                                $product->update_meta_data('rank_math_description', $swProduct->article_metadescription);
                            }
                            // $product->update_meta_data('_yoast_wpseo_title', $swProduct->article_metatitle);
                            // if(!empty($swProduct->article_metakeywords)){
                            //     $product->update_meta_data('_yoast_wpseo_focuskw', $swProduct->article_metakeywords);
                            // }
                            // $product->update_meta_data('_yoast_wpseo_metadesc', $swProduct->article_metadescription);
                        }

                        // Crossell
                        $crossell = [];
                        if(!empty($swProduct->article_crossell)){
                            foreach($swProduct->article_crossell as $a){
                                (!empty($productIDs[$a]) ? $crossell[] = $productIDs[$a] : '');
                            }
                        }
                        $product->set_cross_sell_ids($crossell);

                        // Upsell
                        $upsell = [];
                        if(!empty($swProduct->article_upsell)){
                            foreach($swProduct->article_upsell as $a){
                                (!empty($productIDs[$a]) ? $upsell[] = $productIDs[$a] : '');
                            }
                        }
                        $product->set_upsell_ids($upsell);

                        if(!isset($swretail_options['no_sync']['attributes']) || $swretail_options['no_sync']['attributes'] != 1){

                            // Set weight
                            if(isset($swProduct->article_weight)) {
                                $product->set_weight($swProduct->article_weight);
                            }

                            $productAttributes = [];
                            $productSizeList = [];
                            foreach ($productAttributeMapping as $attribute_key => $attribute_taxonomy) {
                                if (empty($attribute_taxonomy)) {
                                    continue;
                                }

                                if($attribute_taxonomy == '_length'){
                                    $product->set_length( $swProduct->fields->{$attribute_key} );
                                }
                                if($attribute_taxonomy == '_width'){
                                    $product->set_width( $swProduct->fields->{$attribute_key} );
                                }
                                if($attribute_taxonomy == '_height'){
                                    $product->set_height( $swProduct->fields->{$attribute_key} );
                                }

                                $attribute_taxonomy = explode(':', $attribute_taxonomy);
                                $options = [];
                                $isVariation = 0;
                                switch ($attribute_key) {
                                    case 'size':
                                        $options = $productSizeList;
                                        $isVariation = 1;
                                        break;
                                    case 'color':
                                        if(isset($swretail_combine_attributes[$attribute_taxonomy[1]]) && !empty($swProduct->article_color)){
                                            foreach($swretail_combine_attributes[$attribute_taxonomy[1]] as $key => $value){
                                                if(strpos($value,$swProduct->article_color) !== false){
                                                    $options = [$key];
                                                }else{
                                                    $options = [$swProduct->article_color];
                                                }
                                            }
                                        }else{
                                            $options = [$swProduct->article_color];
                                        }
                                        break;
                                    case 'freefield1':
                                        $options = [$swProduct->article_freefield1];
                                        break;
                                    case 'freefield2':
                                        $options = [$swProduct->article_freefield2];
                                        break;
                                    case 'brand':
                                        $options = [$swProduct->article_manufacturer];
                                        break;
                                    default:
                                        if (!empty($swProduct->fields->{$attribute_key})) {
                                            
                                            if(isset($swretail_combine_attributes[$attribute_taxonomy[1]])){
                                                $option_found = false;
                                                foreach($swretail_combine_attributes[$attribute_taxonomy[1]] as $key => $value){
                                                    if(strpos($value,$swProduct->fields->{$attribute_key}) !== false){
                                                        $term = get_term_by('slug', $key, $attribute_taxonomy[1]);
                                                        if($term){
                                                            $options = [$term->name];
                                                            $option_found = true;
                                                        }
                                                    }
                                                    if(!$option_found){
                                                        $options = explode('', $swProduct->fields->{$attribute_key});
                                                    }
                                                }
                                            }else{
                                                $options = explode('', $swProduct->fields->{$attribute_key});
                                            }

                                        }
                                }
                                if (!empty($options)) {
                                    $attribute = new WC_Product_Attribute();
                                    $attribute->set_id($attribute_taxonomy[0]);
                                    $attribute->set_name($attribute_taxonomy[1]);
                                    $attribute->set_visible(1);
                                    $attribute->set_variation($isVariation);
                                    $attribute->set_options($options);
                                    $productAttributes[] = $attribute;
                                }
                            }

                            // Set attributes
                            $product->set_attributes($productAttributes);
                        }

                        if(!isset($swretail_options['no_sync']['categories']) || $swretail_options['no_sync']['categories'] != 1){
                            // Set categories
                            $categoryIds = [];
                            if(!empty($swProduct->article_group)){
                                $category = $this->categoryFindOrCreate($swProduct->article_group);
                                $categoryIds[] = $category;
                                if(!empty($swProduct->article_subgroup)){
                                    $subCategory = $this->categoryFindOrCreate($swProduct->article_subgroup, $category);
                                    $categoryIds[] = $subCategory;
                                    if (!empty($swProduct->article_subsubgroup)) {
                                        $subSubCategory = $this->categoryFindOrCreate($swProduct->article_subsubgroup, $subCategory);
                                        $categoryIds[] = $subSubCategory;
                                    }
                                }
                            }
                            // Extra category groups
                            if(!empty($swProduct->article_group_extra)) {
                                foreach ($swProduct->article_group_extra as $extra_categories) {
                                    if (!empty($extra_categories[0])) {
                                        $category = $this->categoryFindOrCreate($extra_categories[0]);
                                        $categoryIds[] = $category;
                                        if (!empty($extra_categories[1])) {
                                            $subCategory = $this->categoryFindOrCreate($extra_categories[1], $category);
                                            $categoryIds[] = $subCategory;
                                            if (!empty($extra_categories[2])) {
                                                $subSubCategory = $this->categoryFindOrCreate($extra_categories[2], $subCategory);
                                                $categoryIds[] = $subSubCategory;
                                            }
                                        }
                                    }
                                }
                            }

                            // 3.7.0: Add product with a discount price to a specific SALE category
                            if(  isset($swretail_addons['auto_sale_cat']['active']) && $swretail_addons['auto_sale_cat']['active'] == 1 && isset($swretail_options['auto_sale_cat']) ){
                                if($swProduct->article_discount){
                                    $categoryIds[] = $swretail_options['auto_sale_cat'];
                                }elseif($swretail_options['webprice'] == 'discount' && $swProduct->article_price_web && $swProduct->article_price_web < $swProduct->article_baseprice){
                                    $categoryIds[] = $swretail_options['auto_sale_cat'];
                                }
                            }

                            if(!empty($categoryIds)){
                                $product->set_category_ids(array_unique($categoryIds));
                            }
                        }

                        $product->save();
                        
                        if(!isset($swretail_options['no_sync']['images']) || $swretail_options['no_sync']['images'] != 1){
                            //Find or create images
                            $images = $this->imagesFindOrCreate($swProduct->images, $product->get_id());
                            $product->set_image_id($images->mainImage);
                            $product->set_gallery_image_ids($images->galleryImages);
                            $product->save();
                        }
                    }
//                }elseif(!empty($productID) && isset($swProduct->article_inwebshop) && $swProduct->article_inwebshop == 0){
                }elseif(isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1) {

                    // Sync also when addon "syncbybarcodes" is active and "in webshop" is not selected

                    $eans = [];
                    if(!empty($swProduct->barcodes)){
                        foreach ($swProduct->barcodes as $barcode){
                            $eans[] = $barcode->barcode;
                        }
                    }
                    $productID = $this->getWcProductIdBySwBarcode($eans);                     
                    if( $productID != null && (!isset($swretail_options['no_sync']['stock']) || $swretail_options['no_sync']['stock'] != 1)){        
                        
                        $product = new WC_Product($productID);
                        $product->set_manage_stock(1); // 0 = off, 1 = on
                        if($backorders) {
                            $product->set_backorders($backorders); //Options: 'yes', 'no' or 'notify'.
                        }else{
                            if(isset($swProduct->article_run_out)){
                                if($swProduct->article_run_out == 1){
                                    // if($backorders == 'yes') {
                                    //     $product->set_backorders('yes');
                                    // }elseif($backorders == 'no'){
                                        $product->set_backorders('no');
                                    // }
                                }else{
                                    if($swretail_options['backorders'] == 'false-notify'){
                                        $product->set_backorders('notify');
                                    }else{
                                        $product->set_backorders('yes');
                                    }
                                }
                            }
                        }
                        $product->set_stock_quantity($swProduct->sizes[0]->stock);
                        $product->save();
                    }
                    
                }else{
                    wp_delete_post($productID);
                }
            }catch(WC_Data_Exception $ex){
                $this->logError($swProduct->article_description.': '.$ex->getCode().': '.$ex->getMessage());
                $errors[$swProduct->article_description] = $ex->getCode().': '.$ex->getMessage();
            }
            //break;
        }

        if(!empty($swretail_categories)) {
            update_option('swretail_categories', $swretail_categories);
        }

        if(count($errors) > 0){
            return $errors;
        }
        return false;
    }

    /**
     * Calculate discount
     *
     * @param double $regularPrice
     * @param mixed $rawDiscount
     * @return mixed
     */
    private function getDiscount($regularPrice, $rawDiscount){ 
        $rawDiscount = str_replace([',', ' '],['.', ''], $rawDiscount);

        switch ($rawDiscount){
            case (preg_match("/^([*])(\d*[,.\d]*)/", $rawDiscount, $match) ? true : false): // Starting with *
                $discountPrice = $match[2];
                break;
            case (preg_match("/^(\d*[,.\d]*)([*])/", $rawDiscount, $match) ? true : false): // Ending with *
                $discountPrice = $match[1];
                break;
            case (preg_match("/^([%])(\d*[,.\d]*)/", $rawDiscount, $match) ? true : false): // Starting with %
                $discountPrice = $regularPrice - (($regularPrice / 100) * $match[2]);
                break;
            case (preg_match("/^(\d*[,.\d]*)([%])/", $rawDiscount, $match) ? true : false): // Ending with %
                $discountPrice = $regularPrice - (($regularPrice / 100) * $match[1]);
                break;
            default:
                $discountPrice = $regularPrice - $rawDiscount;
        }
        return round($discountPrice,2);
    }

    /**
     * Finds or creates category term and returns term id.
     *
     * @param $name
     * @param int $parent_id
     * @return mixed
     */
    private function categoryFindOrCreate($name, $parent_id = 0){
        $categories = get_terms([
            'taxonomy' => 'product_cat',
            'name'	=> $name,
            'hide_empty' => false,
            'parent' => $parent_id
        ]);
        if(empty($categories)) {
            $newCategory = wp_insert_term($name, 'product_cat', ['parent' => $parent_id]);
            return $newCategory['term_id'];
        }else{
            return $categories[0]->term_id;
        }
    }

    /**
     * Find images or create/upload new images by filename.
     *
     * @param object $swImages
     * @param int $productID
     * @return object
     */
    private function imagesFindOrCreate($swImages, $productID){
        global $wpdb;
        $swretail_options = get_option('swretail_options');
        $productImageIds = [];
        $useSubFolders = get_option('uploads_use_yearmonth_folders');

        $newChecksum = '';
        foreach($swImages as $swImage){
            if($swImage->image_checksum){
                $newChecksum .= $swImage->image_checksum;
            }
        }

        $forceNewImages = false;
        if( get_post_meta($productID,'image_checksum',true) ){
            $checksum = get_post_meta($productID,'image_checksum',true);
            if($checksum != $newChecksum){
                $forceNewImages = true;
                update_post_meta($productID,'image_checksum',$newChecksum);
            }
        }else{
            $forceNewImages = true;
            update_post_meta($productID,'image_checksum',$newChecksum);
        }

        // Set forceNewImages to true for debugging
        // $forceNewImages = true;
        foreach($swImages as $swImage){

            $imageFileName = $swImage->image_file;
            $imageExtension = end(explode('.',$swImage->image_file));
            if(isset($swretail_options['description_as_filename']) && $swretail_options['description_as_filename'] == 'yes' && !empty($swImage->image_description)){
                $imageFileName = sw_clean_url($swImage->image_description).'.'.$imageExtension;
            }
            
            $fullImageFileName = $swImage->image_checksum.'_'.$imageFileName;
            if(isset($swretail_options['remove_checksum']) && $swretail_options['remove_checksum'] == 'yes'){
                $fullImageFileName = $imageFileName;
            }         
            // Remove file extension because Wordpress adds "scaled" to large images.
            $fullImageFileNameNoExt = str_replace($imageExtension,'',$fullImageFileName);

            if($useSubFolders == 1){
                $image = $wpdb->get_row('SELECT post_id FROM '.$wpdb->postmeta.' WHERE meta_key = "_wp_attached_file" AND meta_value LIKE "%/'.$fullImageFileNameNoExt.'%"');
            }else{
                $image = $wpdb->get_row('SELECT post_id FROM '.$wpdb->postmeta.' WHERE meta_key = "_wp_attached_file" AND meta_value LIKE "%'.$fullImageFileNameNoExt.'%"');
            }
            if($image && !$forceNewImages){
                $productImageIds[] = $image->post_id;
            }else{
                if($image && $forceNewImages){
                    wp_delete_attachment($image->post_id,true);
                }
                // Upload new image
                if(!empty($swImage->image_file) && $this->getSwImage($swImage->image_file,$fullImageFileName,$swImage->image_checksum)){
                    $productImageIds[] = media_sideload_image(plugin_dir_url( __FILE__ ).'tmp/'.$fullImageFileName, $productID, $swImage->image_description, 'id');
                    unlink(plugin_dir_path( __FILE__ ).'tmp/'.$fullImageFileName);
                }
            }
        }
        $return = new stdClass();
        $return->mainImage = reset($productImageIds);
        
        if(isset($swretail_options['featured_image_in_gallery']) && $swretail_options['featured_image_in_gallery'] == 'yes'){
            $return->galleryImages = $productImageIds;
        }else{
            $return->galleryImages = array_slice($productImageIds, 1);
        }

        return $return;
    }

    /**
     * Download file to tmp directory, returns true if success else false.
     *
     * @param string $swfilename
     * @return bool
     */
    private function getSwImage($swfilename,$filename,$checksum){
        try{
            swretailUpdateApiRequestCount('article_image');
            $response = wp_remote_get( $this->swURL.'article_image/'.$swfilename, $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getSwImage()');
            }else{
                file_put_contents(plugin_dir_path( __FILE__ ).'tmp/'.$filename, $response['body']);
                return true;
            }
        }catch(Exception $ex){
            $this->logError($ex);
            return false;
        }
    }

    /**
     * Update or create SW-Retail Relation
     * 
     * @param array $customer
     * @return string relationcode
     */
    public function swRelationUpdateOrCreate($customer){
//		if(!$this->checkLicence()){return;};
        try{
            $unifiedzip = strtoupper(str_replace(' ', '', $customer['zipcode'].$customer['housenumber']));
            swretailUpdateApiRequestCount('relation/unifiedzip');
            $response = wp_remote_get( $this->swURL.'relation/unifiedzip/'.$unifiedzip, $this->wpRemoteArgs );
            if(is_wp_error($response)){
                swretailUpdateApiRequestCount('relation/email');
                $response = wp_remote_get( $this->swURL.'relation/email/'.$customer['email'], $this->wpRemoteArgs );
            }
            // return $this->swURL.'relation/unifiedzip/'.$unifiedzip;
            // return $response;

            //Check for WP errors
            if(is_wp_error($response)){
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', swRelationUpdateOrCreate()');
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode) || $response->errorcode == 0){

                    // 2023-05-02 - $response can be an array of SW Retail relations
                    if(is_array($response)){
                        $response = $response[0];
                    }

                    // return $response;

                    $relation = [
                        'lastname'			=> $customer['lastname'],
                        'firstname'			=> $customer['firstname'],
                        'street'			=> $customer['street'],
                        'city'				=> $customer['city'],
                        'housenumber'		=> $customer['housenumber'],
                        'zipcode'			=> $customer['zipcode'],
                        'phone1'			=> $customer['phone1'],
                        'phone2'			=> $customer['phone2'],
                        'country'			=> $customer['country'],
                        'email'				=> $customer['email'],
                        'relationtype'		=> 2, //1=supplier, 2=customer, 4=manufacturer
                        'sex'				=> '',
                        'external_id'		=> '',
                    ];

                    if(!empty($customer['contact'])){
                        $relation['contact'] = $customer['contact'];
                    }

                    if(!empty($customer['external_id'])){
                        $relation['external_id'] = $customer['external_id'];
                    }

                    //If logedin customer check external_id
                    if(empty($response->relationcode) && !empty($customer['external_id'])){
                        $secondaryResponse = wp_remote_get( $this->swURL.'relation/external_id/'.$customer['external_id'], $this->wpRemoteArgs );
                        if(is_wp_error($secondaryResponse)){
                            //throw new Exception('WP error: #'.$secondaryResponse->get_error_code().' - '.$secondaryResponse->get_error_message().', swRelationUpdateOrCreate()');
                        }else{
                            $secondaryResponse = json_decode($secondaryResponse['body']);
                            //Check for SW-Retail errors
                            if(!isset($secondaryResponse->errorcode) || $secondaryResponse->errorcode == 0){
                                if(isset($response->relationcode)){
                                    $relation['relationcode'] = $response->relationcode;
                                }
                            }
                        }
                    }elseif($response->relationcode){
                        $relation['relationcode'] = $response->relationcode;
                    }
                    $post = $this->wpRemoteArgs;
                    $post['headers']['Content-Type'] = 'application/json';
                    $post['body'] = json_encode($relation);
                    swretailUpdateApiRequestCount('relation');
                    $postResponse = wp_remote_post( $this->swURL.'relation', $post );

                    //Check for WP errors
                    if(is_wp_error($postResponse)){
                        throw new Exception('WP error: #'.$postResponse->get_error_code().' - '.$postResponse->get_error_message().', swRelationUpdateOrCreate() swRelationUpdateOrCreate() -> wp_remote_post()');
                    }else{
                        $postResponse = json_decode($postResponse['body']);
                        //Check for SW-Retail errors
                        if(!isset($postResponse->errorcode)){
                            return $postResponse->relationcode;
                        }else{
                            throw new Exception('SW-Retail error: #'.$postResponse->errorcode.' - '.$postResponse->errorstring.', swRelationUpdateOrCreate() -> wp_remote_post(');
                        }
                    }
                    
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', swRelationUpdateOrCreate()');
                }
            } 
        }catch(Exception $ex){
//			echo $ex;
            $this->logError($ex);
            return false;
        }
    }

    /**
     * Post WooCommerce order to SW-Retail
     *
     * @param WC_Order $wcOrder
     * @param $attempt
     * @return boolean $postResponse true on success or false on error.
     */
    public function postSwOrder(WC_Order $wcOrder, $attempt = 1){
        global $woocommerce;
        $swretail_options = get_option('swretail_options');
        $swretail_addons = get_option('swretail_addons');
		if(!$this->checkLicence()){return;};

        $sw_use_prod_bundles = false;
        if( isset($swretail_addons['wc_product_bundles']['active']) ){
            $swRetail = new SWRetail($swretail_options['licence_key'], 'https://'.$swretail_options['api_url'].'/swcloud/SWWService/', $swretail_options['api_user'], $swretail_options['api_pass']);
            if($swRetail->checkAddon('wc_product_bundles')){
                $sw_use_prod_bundles = true;
            }
        }

        if($wcOrder->get_meta('wwp_wholesale_role') == 'wholesale_customer'){
            $billing = [
                'contact'           => $wcOrder->get_billing_first_name().' '.$wcOrder->get_billing_last_name(),
                'lastname'			=> $wcOrder->get_billing_company(),
                'firstname'			=> $wcOrder->get_billing_first_name(),
            ];
        }else{
            $billing = [
                'lastname'			=> $wcOrder->get_billing_last_name(),
                'firstname'			=> $wcOrder->get_billing_first_name(),
            ];
        }

        $billing_housenumber = (isset($swretail_options['billing_house_number_field']) ? get_post_meta( $wcOrder->get_id(), $swretail_options['billing_house_number_field'], true ) : '');
        $billing_housenumber_extension = (isset($swretail_options['billing_house_number_extension_field']) && !empty($swretail_options['billing_house_number_extension_field']) ? get_post_meta( $wcOrder->get_id(), $swretail_options['billing_house_number_extension_field'], true ) : '');

        $billing = array_merge($billing, [
            // 'street'			=> str_replace($billing_housenumber.' '.$billing_housenumber_extension,'',$wcOrder->get_billing_address_1()),
            'street'			=> str_replace($billing_housenumber.(!empty($billing_housenumber_extension) ? ' '.$billing_housenumber_extension : ''),'',$wcOrder->get_billing_address_1()),
            'city'				=> $wcOrder->get_billing_city(),
            'housenumber'		=> (!empty($billing_housenumber) ? $billing_housenumber.' '.$billing_housenumber_extension : $wcOrder->get_billing_address_2()),
            'zipcode'			=> $wcOrder->get_billing_postcode(),
            'phone1'			=> $wcOrder->get_billing_phone(),
            'email'				=> $wcOrder->get_billing_email(),
            'phone2'			=> '',
            'country'			=> $wcOrder->get_billing_country(),
        ]);


        if(!empty($wcOrder->get_customer_id())){
            $billing['external_id'] = $wcOrder->get_customer_id();
        }

        //Update or create billing relation
        $swBillingRelationId = $this->swRelationUpdateOrCreate($billing);

        if(strtoupper($wcOrder->get_billing_postcode()) == strtoupper($wcOrder->get_shipping_postcode())){ //Billing address is the same as shipping address
            $swShippingRelationId = $swBillingRelationId;
        }else{
            if($wcOrder->get_meta('wwp_wholesale_role') == 'wholesale_customer'){
                $shipping = [
                    'contact'			=> $wcOrder->get_shipping_first_name().' '.$wcOrder->get_shipping_last_name(),
                    'lastname'			=> $wcOrder->get_billing_company(),
                    'firstname'			=> $wcOrder->get_shipping_first_name(),
                ];
            }else{
                $shipping = [
                    'lastname'			=> $wcOrder->get_shipping_last_name(),
                    'firstname'			=> $wcOrder->get_shipping_first_name(),
                ];
            }

            $shipping_housenumber = (isset($swretail_options['shipping_house_number_field']) ? get_post_meta( $wcOrder->get_id(), $swretail_options['shipping_house_number_field'], true ) : '');
            $shipping_housenumber_extension = (isset($swretail_options['shipping_house_number_extension_field']) && !empty($swretail_options['shipping_house_number_extension_field']) ? get_post_meta( $wcOrder->get_id(), $swretail_options['shipping_house_number_extension_field'], true ) : '');
        
            $shipping = array_merge($shipping, [
                // 'street'			=> str_replace($shipping_housenumber.' '.$shipping_housenumber_extension,'',$wcOrder->get_shipping_address_1()),
                'street'			=> str_replace($shipping_housenumber.(!empty($shipping_housenumber_extension) ? ' '.$shipping_housenumber_extension : ''),'',$wcOrder->get_shipping_address_1()),
                'city'				=> $wcOrder->get_shipping_city(),
                'housenumber'		=> (!empty($shipping_housenumber) ? $shipping_housenumber.' '.$shipping_housenumber_extension : $wcOrder->get_shipping_address_2()),
                'zipcode'			=> $wcOrder->get_shipping_postcode(),
                'phone1'			=> '',
                'phone2'			=> '',
                'country'			=> $wcOrder->get_shipping_country(),
            ]);

            if(!empty($wcOrder->get_customer_id())){
                $shipping['external_id'] = $wcOrder->get_customer_id();
            }

            //Update or create shipping relation
            $swShippingRelationId = $this->swRelationUpdateOrCreate($shipping);
        }

        $inetnumber = $wcOrder->get_id();

        if(isset($swretail_options['order_id_meta']) && !empty($swretail_options['order_id_meta'])){
            $inetnumber = get_post_meta($wcOrder->get_id(), $swretail_options['order_id_meta'], true); 
        }

        if(isset($swretail_addons['effectconnect']['active']) && $swretail_addons['effectconnect']['active'] == 1){
            if(isset($swretail_addons['effectconnect']['order_id']) && $swretail_addons['effectconnect']['order_id'] != 'default'){
                if(get_post_meta($wcOrder->get_id(),$swretail_addons['effectconnect']['order_id'],true)) {
                    $inetnumber = get_post_meta($wcOrder->get_id(), $swretail_addons['effectconnect']['order_id'], true);
                }
            }
        } 

        $swOrderData = [
            'inetnumber' 			=> $inetnumber,
            'web_state' 			=> $wcOrder->get_status(),
            'remark' 				=> $wcOrder->get_customer_note(),
            'shipper' 				=> $wcOrder->get_shipping_method(),
            'tracker'				=> '',
            'date' 					=> date('Y-m-d', strtotime($wcOrder->get_date_created())),
            'time' 					=> date('H:i', strtotime($wcOrder->get_date_created())),
            'relation_code_ship' 	=> $swShippingRelationId,
            'relation_code_invoice' => $swBillingRelationId,
            'payment_method' 		=> $wcOrder->get_payment_method_title(),
            'payment_status' 		=> ($wcOrder->get_payment_method_title() == 'iDEAL' ? 'Success' : ''),
            'shipping_value' 		=> $wcOrder->get_shipping_total(),
            'order_lines'			=> [],
        ];

        if($wcOrder->get_meta('wwp_wholesale_role') == 'wholesale_customer'){
            $swOrderData['wholesale'] = 1;
            $swOrderData['shipping_value'] = ($wcOrder->get_shipping_total() - ($swOrderData['wholesale'] == 1 ? 0 : $swOrderData->get_shipping_tax()));
        }

        $coupon = false;
        $couponDiscount = 0;
        if(isset($swretail_options['coupon_id']) && !empty($swretail_options['coupon_id'])){
            if(!empty($wcOrder->get_coupon_codes())){
                $coupon = true;
            }
        }
		
        foreach ($wcOrder->get_items() as $wcOrderItem){

            $swArticleId = get_post_meta($wcOrderItem->get_product_id(), '_sw_article_id', true);
            $taxClass = $wcOrderItem->get_tax_class();
            if(!empty($swretail_options['tax_rate_reduced']) && $taxClass == sanitize_title_with_dashes($swretail_options['tax_rate_reduced'])){
                $taxRate = 9;
            }elseif (!empty($swretail_options['tax_rate_default']) && $taxClass == sanitize_title_with_dashes($swretail_options['tax_rate_default'])){
                $taxRate = 21;
            }else{
                $taxRate = 21;
            }

            if(get_post_meta($wcOrderItem->get_product_id(), '_sw_salesunit', true)){
			    $quantity = $wcOrderItem->get_quantity() * get_post_meta($wcOrderItem->get_product_id(), '_sw_salesunit', true);
            }else{
				$quantity = $wcOrderItem->get_quantity();
            }

            if($wcOrderItem->get_variation_id() != 0){
				
                $swSizePosition = get_post_meta($wcOrderItem->get_variation_id(), '_sw_size_pos', true);
                if(isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1){
                    $variation = new WC_Product_Variation($wcOrderItem->get_variation_id());
                    $linetotal = ($wcOrderItem->get_total() + (isset($swOrderData['wholesale']) && $swOrderData['wholesale'] == 1 ? 0 : $wcOrderItem->get_total_tax()));
                    if($coupon){
                        $linetotal = $wcOrderItem->get_subtotal() + $wcOrderItem->get_subtotal_tax();
                        $discountPrice = $wcOrderItem->get_total() + $wcOrderItem->get_total_tax();
                        $couponDiscount = $couponDiscount + ($linetotal-$discountPrice);
                    }
                    $swOrderData['order_lines'][] = [
                        'article_barcode' => $variation->get_sku(),
                        'discount' => 0,
                        'linetotal' => sw_calculate_price($linetotal,'',true),
                        'taxrate' => $taxRate,
                        'amount' => $quantity,
                        'size_pos' => $swSizePosition,
                    ];
                }else {
                    $linetotal = ($wcOrderItem->get_total() + (isset($swOrderData['wholesale']) && $swOrderData['wholesale'] == 1 ? 0 : $wcOrderItem->get_total_tax()));
                    if($coupon){
                        $linetotal = $wcOrderItem->get_subtotal() + $wcOrderItem->get_subtotal_tax();
                        $discountPrice = $wcOrderItem->get_total() + $wcOrderItem->get_total_tax();
                        $couponDiscount = $couponDiscount + ($linetotal-$discountPrice);
                    }
                    $swOrderData['order_lines'][] = [
                        'article_id' => $swArticleId,
                        'discount' => 0,
                        'linetotal' => sw_calculate_price($linetotal,'',true),
                        'taxrate' => $taxRate,
                        'amount' => $quantity,
                        'size_pos' => $swSizePosition,
                    ];
                }
            }elseif(wc_get_order_item_meta($wcOrderItem->get_id(), '_bundle_cart_key', true ) && $sw_use_prod_bundles ){
                // WooCommmerce Product Bundles
                // Integrated in SW Retail for Woocommerce since 2022-09-12
                
                $sw_prod_bundle_id = (!empty($swretail_addons['wc_product_bundles']['sw_prod_id']) ? $swretail_addons['wc_product_bundles']['sw_prod_id'] : 0 );
				
                if(wc_get_order_item_meta($wcOrderItem->get_id(), '_bundled_by', true )){
					
					if(isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1) {
						if($_subproduct = wc_get_product($wcOrderItem->get_product_id())) {

                            if(isset($swretail_options['barcode_meta']) && !empty($swretail_options['barcode_meta'])){
                                $sku = get_post_meta($wcOrderItem->get_product_id(),$swretail_options['barcode_meta'],true);
                            }else{
                                $sku = $_subproduct->get_sku();
                            }

							$swOrderData['order_lines'][] = [
                        		'article_barcode' => $sku,
								'discount' => $_subproduct->get_regular_price(),
								'linetotal' => 0,
								'taxrate' => 0,
								'amount' => $quantity,
								'size_pos' => 1
							];
						}						
					}else{
						if($_subproduct = wc_get_product($wcOrderItem->get_product_id())) {
							$swOrderData['order_lines'][] = [
								'article_id' => $swArticleId,
								'discount' => $_subproduct->get_regular_price(),
								'linetotal' => 0,
								'taxrate' => 0,
								'amount' => $quantity,
								'size_pos' => 1
							];
						}
					}					

                }else{
                    $linetotal = ($wcOrderItem->get_total() + (isset($swOrderData['wholesale']) && $swOrderData['wholesale'] == 1 ? 0 : $wcOrderItem->get_total_tax()));
                    if($coupon){
                        $linetotal = $wcOrderItem->get_subtotal() + $wcOrderItem->get_subtotal_tax();
                        $discountPrice = $wcOrderItem->get_total() + $wcOrderItem->get_total_tax();
                        $couponDiscount = $couponDiscount + ($linetotal-$discountPrice);
                    }
                    $swOrderData['order_lines'][] = [
                        'article_id'	=> $sw_prod_bundle_id,
                        'discount'		=> 0,
                        'linetotal'		=> sw_calculate_price($linetotal,'',true),
                        'taxrate'		=> $taxRate,
                        'amount'		=> $quantity,
                        'size_pos'		=> 1
                    ];
                }

            }else{
				
                if(isset($swretail_addons['syncbybarcodes']['active']) && $swretail_addons['syncbybarcodes']['active'] == 1) {
                    $product = new WC_Product($wcOrderItem->get_product_id());		
                    
                    if(isset($swretail_options['barcode_meta']) && !empty($swretail_options['barcode_meta'])){
                        $sku = get_post_meta($wcOrderItem->get_product_id(),$swretail_options['barcode_meta'],true);
                    }else{
                        $sku = $product->get_sku();
                    }
					
                    $linetotal = ($wcOrderItem->get_total() + (isset($swOrderData['wholesale']) && $swOrderData['wholesale'] == 1 ? 0 : $wcOrderItem->get_total_tax()));
                    if($coupon){
                        $linetotal = $wcOrderItem->get_subtotal() + $wcOrderItem->get_subtotal_tax();
                        $discountPrice = $wcOrderItem->get_total() + $wcOrderItem->get_total_tax();
                        $couponDiscount = $couponDiscount + ($linetotal-$discountPrice);
                    }
                    $swOrderData['order_lines'][] = [
                        'article_barcode' => $sku,
                        'discount' => 0,
                        'linetotal' => sw_calculate_price($linetotal,'',true),
                        'taxrate' => $taxRate,
                        'amount' => $quantity,
                        'size_pos' => 1
                    ];
                }else{
                    $linetotal = ($wcOrderItem->get_total() + (isset($swOrderData['wholesale']) && $swOrderData['wholesale'] == 1 ? 0 : $wcOrderItem->get_total_tax()));
                    if($coupon){
                        $linetotal = $wcOrderItem->get_subtotal() + $wcOrderItem->get_subtotal_tax();
                        $discountPrice = $wcOrderItem->get_total() + $wcOrderItem->get_total_tax();
                        $couponDiscount = $couponDiscount + ($linetotal-$discountPrice);
                    }
                    $swOrderData['order_lines'][] = [
                        'article_id' => $swArticleId,
                        'discount' => 0,
                        'linetotal' => sw_calculate_price($linetotal,'',true),
                        'taxrate' => $taxRate,
                        'amount' => $quantity,
                        'size_pos' => 1
                    ];
                }
            }
        }

        if($coupon){
            if($couponDiscount > 0){
                $swOrderData['order_lines'][] = [
                    'article_id' => $swretail_options['coupon_id'],
                    'discount' => 0,
                    'linetotal' => sw_calculate_price(($couponDiscount*-1),true),
                    'taxrate' => 0,
                    'amount' => 1,
                    'size_pos' => 1
                ];
            }else{
                $swOrderData['order_lines'][] = [
                    'article_id' => $swretail_options['coupon_id'],
                    'discount' => 0,
                    'linetotal' => sw_calculate_price($couponDiscount),
                    'taxrate' => 0,
                    'amount' => 1,
                    'size_pos' => 1
                ];
            }
        }

        try{
            $post = $this->wpRemoteArgs;
            $post['headers']['Content-Type'] = 'application/json';
            $post['body'] = json_encode($swOrderData);
            swretailUpdateApiRequestCount('order');
            $postResponse = wp_remote_post( $this->swURL.'order', $post );
            //Check for WP errors

            if(is_wp_error($postResponse)) {
                throw new Exception('WP error: #'.$postResponse->get_error_code().' - '.$postResponse->get_error_message().', postSwOrder() -> wp_remote_post()');
            }else{
                $postResponse = json_decode($postResponse['body']);
                //Check for SW-Retail errors

                if(!isset($postResponse->errorcode)){
                    update_post_meta( $wcOrder->get_id(), '_wc_order_send_to_swretail', date('Y-m-d H:i:s') );
                    $message =  __( 'Order send to SW-Retail', 'swretail' );
                    $wcOrder->add_order_note( $message );
                    return true;
                }else{
                    m2m_sw_slack(serialize($postResponse));
                    throw new Exception('SW-Retail error: #'.$postResponse->errorcode.' - '.$postResponse->errorstring.', postSwOrder() -> wp_remote_post(');
                }
            }
        }catch(Exception $ex){
            $this->logError('Error processing order '.$wcOrder->get_id().':'.$ex->getMessage());
            $message = sprintf(__( 'A error accrued wile trying to send order to SW-Retail (attempt %1$s)', 'swretail'), $attempt);
            $wcOrder->add_order_note( $message );
            if($attempt <= 2){ 
                wp_schedule_single_event( time() + 60, 'swretail_event_process_orders', [$wcOrder->get_id(), $attempt+1]);
            }
            return false;
        }
    }

    /**
     * Get SW-Retail changed order ids
     *
     * @param $minutes
     * @return boolean $postResponse true on success or false on error.
     */
    public function getChangedSwOrders($minutes = 10){
        if(!$this->checkLicence()){return;};
        try{
            $response = wp_remote_get( $this->swURL.'order_changed/'.$minutes, $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getChangedSwOrders()');
            }else{
                $response = json_decode($response['body']);
                //Check for SW-Retail errors
                if(!isset($response->errorcode)){
//					$responseChunks = array_chunk($response, 10);
//					$timeoffset = 1;
                    foreach ($response as $order){
                        $this->getSwOrder($order);
////						$this->createSwTask('update_or_create_products', implode(',', $chunk));
//						wp_schedule_single_event( time() + (60 * $timeoffset), 'swretail_event_process_changes', [implode(',', $chunk)]);
//						$timeoffset++;
                    }
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getChangedSwOrders()');
                }
            }
            return true;
        }catch(Exception $ex){
            // $this->logError($ex->getMessage());            
            m2m_sw_slack($ex->getMessage());
            return false;
        }
    }

    /**
     * Get SW-Retail order
     * @param $id
     */
    public function getSwOrder($id){
        global $wpdb;
        $swretail_options = get_option('swretail_options');

        if(!$this->checkLicence()){return;};
        try{
            $response = wp_remote_get( $this->swURL.'order/'.$id, $this->wpRemoteArgs );
            //Check for WP errors
            if(is_wp_error($response)) {
                throw new Exception('WP error: #'.$response->get_error_code().' - '.$response->get_error_message().', getChangedSwOrders()');
            }else{
                $response = json_decode($response['body']);

                //Check for SW-Retail errors
                if(!isset($response->errorcode)){

                    if(!empty($response->inetnumber)){
                        if(strpos($response->inetnumber, '_C') && !empty($swretail_options['credits']) && $swretail_options['credits'] != 'no'){
                            $refundExists = $wpdb->get_row('SELECT post_id FROM '.$wpdb->postmeta.' WHERE meta_key = "_sw_credit_id" AND meta_value LIKE "'.$response->inetnumber.'"');
                            if(!empty($refundExists)){
                                return;
                            }

                            $wcOrder = new WC_Order(str_replace('_C', '', $response->inetnumber));
                            $refund = [
                                'amount' => 0,
                                'reason' => '',
                                'order_id' => str_replace('_C', '', $response->inetnumber),
                                'line_items' => [],
                                'restock_items' => false,
                            ];
                            foreach ($response->order_lines as $orderLine){
                                if($orderLine->amount < 0) {

                                    if(!empty($orderLine->size_pos)){
                                        $prodId = $this->getWcVariationIdBySwArticleId($orderLine->article_id, $orderLine->size_pos);
                                    }else{
                                        $prodId = $this->getWcProductIdBySwArticleId($orderLine->article_id);
                                    }

                                    foreach ($wcOrder->get_items() as $wcOrderItem){
                                        if($wcOrderItem->get_product_id() == $prodId || $wcOrderItem->get_variation_id() == $prodId){
                                            $lineId = $wcOrderItem->get_id();
                                        }
                                    }

                                    $refund['line_items'][$lineId ?? $orderLine->article_id] = [
                                        'qty' => $orderLine->amount,
                                        'refund_total' => wc_format_decimal($orderLine->linetotal, wc_get_price_decimals()),
                                        'refund_tax' => ($orderLine->linetotal / ($orderLine->taxrate + 100) * $orderLine->taxrate),
                                    ];
                                    $refund['amount'] = ($refund['amount'] + abs($orderLine->linetotal));
                                }
                            }
                            $orderRefund = wc_create_refund($refund);
                            if(is_wp_error($orderRefund)) {
                                throw new Exception('WP error: #'.$orderRefund->get_error_code().' - '.$orderRefund->get_error_message().', getSwOrder()');
                            }
                            $orderRefund->update_meta_data('_sw_credit_id', $response->inetnumber);
                            $orderRefund->save();
                        }elseif(strpos($response->inetnumber, '_R')){
//							$order = new WC_Order(str_replace('_R', '', $response->inetnumber));
//                            $order->update_status('refunded', 'order_note');
                        }else{
                            $order = new WC_Order($response->inetnumber);
                            switch ($response->swretail_state){
//								case 'Retournota':
//									$order->update_status('refunded', 'order_note');
//									break;
                                case 'Creditnota':
//									$order->update_status('refunded', 'order_note');
//									break;
                                case 'Verstuurd':
//									$order->update_status('completed', 'order_note');
//									break;
                            }
                        }

                    }
                }else{
                    throw new Exception('SW-Retail error: #'.$response->errorcode.' - '.$response->errorstring.', getChangedSwOrders()');
                }
            }
        }catch(Exception $ex){
            // $this->logError($ex->getMessage());            
            m2m_sw_slack($ex->getMessage());
        }
    }
}